import os

import numpy as np
import pandas as pd

import pathlib

import skimage.exposure
import skimage.measure

# Add to path submodule to be imported

from yeastmatedetector.inference import YeastMatePredictor

from cellacdc.core import getBaseCca_df

class Model:
    def __init__(self):
        user_path = pathlib.Path.home()
        model_path = os.path.join(str(user_path), f'acdc-YeastMate')
        yaml_path = os.path.join(model_path, 'yeastmate.yaml')
        weights_path = os.path.join(model_path, 'yeastmate_weights.pth')

        self.model = YeastMatePredictor(
            yaml_path,
            weights_path
        )

    def segment(self, image,
                score_thresholds={0:0.9, 1:0.75, 2:0.75},
                pixel_size=110,
                reference_pixel_size=110,
                lower_quantile=1.5,
                upper_quantile=98.5):

        detections, lab = self.model.inference(
            image,
            score_thresholds=score_thresholds,
            pixel_size=pixel_size,
            reference_pixel_size=reference_pixel_size,
            lower_quantile=lower_quantile,
            upper_quantile=upper_quantile
        )
        return lab

    def predictCcaState(self, image, precomputed_lab):
        detections, lab = self.model.inference(image)
        rp = skimage.measure.regionprops(lab)
        IDs = [obj.label for obj in rp]
        areas = [obj.area for obj in rp]
        centroids = [obj.centroid for obj in rp]
        precomputedIDs = [
            obj.label for obj in skimage.measure.regionprops(precomputed_lab)
        ]
        cca_df = getBaseCca_df(precomputedIDs)
        for obj in rp:
            info = detections.get(str(obj.label))
            if info is None:
                continue

            obj_class = info.get('class')
            if len(obj_class) < 2:
                continue

            is_budding = float(obj_class[1])>2
            if not is_budding:
                continue

            links = info.get('links')
            if not links:
                continue

            link = links[0]
            mother_bud_info = detections.get(link)
            if mother_bud_info is None:
                continue

            mother_bud_ids = mother_bud_info.get('links')
            if mother_bud_ids is None:
                continue

            ID1, ID2 = int(mother_bud_ids[0]), int(mother_bud_ids[1])
            idx1 = IDs.index(ID1)
            idx2 = IDs.index(ID2)
            area1 = areas[IDs.index(ID1)]
            area2 = areas[IDs.index(ID2)]
            if area1 > area2:
                moth_y, moth_x = centroids[idx1]
                bud_y, bud_x = centroids[idx2]
            else:
                moth_y, moth_x = centroids[idx2]
                bud_y, bud_x = centroids[idx1]

            mothID = precomputed_lab[int(round(moth_y)), int(round(moth_x))]
            budID = precomputed_lab[int(round(bud_y)), int(round(bud_x))]

            if mothID not in cca_df.index:
                continue

            if budID not in cca_df.index:
                continue

            cca_df.at[mothID, 'relative_ID'] = budID
            cca_df.at[mothID, 'cell_cycle_stage'] = 'S'

            cca_df.at[budID, 'relative_ID'] = mothID
            cca_df.at[budID, 'cell_cycle_stage'] = 'S'
            cca_df.at[budID, 'relationship'] = 'bud'
            cca_df.at[budID, 'generation_num'] = 0
        return cca_df

def url_help():
    return 'https://github.com/hoerlteam/YeastMate/blob/main/examples/python_detection.ipynb'



import numpy as np
import os
from PIL import Image
import scipy.io as sio
from nd2reader import ND2Reader

try:
    from . import myutils
except Exception as e:
    import myutils

### Helper Functions

def get_metadata_from_nd2(path, drop_z_coords=True):
    """
    function to extract the meta data of a .nd2 file.
    Reads nd2 file, extracts meta data and returns it in form of a dictionary
    Parameters
    ----------
    path: String
        path of the .nd2 file
    drop_z_coords: bool
        Standard True, set to False if convocal data is used
    Returns
    -------
    meta_data: dict
        dictionary containing the meta data of the .nd2 file
    """
    with ND2Reader(path) as images:
        meta_data = images.metadata
        if drop_z_coords:
            del meta_data['z_coordinates']
        return meta_data


def get_index_from_pos_ch_t(pos, ch, t, meta_data, parse_from_single_pos_file=False):
    """
    function to calculate the ND2Reader index of a particular frame.
    The function get_frame_2D does (afaik) not
    work for non-convocal data. A translation of position, channel,
    and time frame into the one-dimensional index of the flattened
    data is therefore needed and performed in this function.
    Parameters
    ----------
    pos: int
        position (/field of view) of the frame which should be extracted
    ch: int
        Channel of the frame which should be extracted
    t: int
        frame id of the frame which should be extracted
    meta_data: dict
        dictionary containing the .nd2 file's meta data as extracted
        by the function get_metadata_from_nd2()
    parse_from_single_pos_file: bool
        if True, the number of positions from the meta data is ignored.
    Returns
    -------
    idx: int
        index of the frame which can be used in parser.get_image()
        of the module ND2Reader. This index identifies the frame
        with the configured position, channel, and point of time.
    """
    n_pos = len(meta_data['fields_of_view'])
    n_channels = len(meta_data['channels'])
    n_frames = meta_data['num_frames']
    if parse_from_single_pos_file:
        idx = t * n_channels + ch
    else:
        idx = t * (n_pos * n_channels) + pos * n_channels + ch
    return idx


### Data handlers


def get_npy_from_nd2(path, positions='all', channels='all', frames='all',\
                     cut_images=0, convocal=False, parse_from_single_pos_file=False):
    """
    function to load the specified frames into a numpy array.
    The function takes a .nd2 file's path and ranges of positions,
    channels, and frames and returns a numpy array containing the
    image data of the specified input. In the course of loading the
    data, an manual cut can also be performed to save disk space
    and crop the data to relevant regions.
    Parameters
    ----------
    path: str
        path of the .nd2 file
    positions: list or str 'all'
        default 'all'. List of positions which should be extracted
        from the .nd2 file. If set to 'all', all available positions
        are extracted.
    channels: list or str 'all'
        default 'all'. List of channels which should be extracted
        from the .nd2 file. If set to 'all', all available channels
        are extracted.
    frames: list or str 'all'
        default 'all'. List of time frames which should be extracted
        from the .nd2 file. If set to 'all', all available frames
        are extracted.
    cut_images: int
        default 0. Number of pixels which are cropped away at both sides
        of the x and y dimension.
    convocal: bool
        default False. Specifying if the data is convocal. This is
        needed to extract the correct index in the flattened data
        using the function get_index_from_pos_ch_t().
    Returns
    -------
    img_array: np.array
        numpy array of dimension(pos, ch, t, x, y) containing the image
        data of the specified positions, channels and frames.
    """
    # only keep z_coords in case of convocal data
    drop_z_coords = bool(1-convocal)
    meta_data = get_metadata_from_nd2(path, drop_z_coords=drop_z_coords)
    # translate position, channel and frames arguments into lists
    if positions == 'all':
        pos_list = list(meta_data['fields_of_view'])
    else:
        pos_list = positions
    if channels == 'all':
        ch_list = list(range(len(meta_data['channels'])))
    else:
        ch_list = channels
    if frames == 'all':
        t_list = list(range(meta_data['num_frames']))
    else:
        t_list = frames
    # build image_array
    ## determine dimensions of final array
    if cut_images==0:
        x_dim, y_dim = meta_data['width'], meta_data['height']
    else:
        x_dim = meta_data['width'] - 2*cut_images
        y_dim = meta_data['height'] - 2*cut_images
    img_array_shape = (len(pos_list), len(ch_list), len(t_list), y_dim, x_dim)
    img_array = np.zeros(img_array_shape)
    ## load each position, channel, frame and save to array
    with ND2Reader(path) as images:
        parser = images.parser
        for pos_idx in range(len(pos_list)):
            for ch_idx in range(len(ch_list)):
                for t_idx in range(len(t_list)):
                    img_idx = get_index_from_pos_ch_t(pos_list[pos_idx], ch_list[ch_idx],\
                                                      t_list[t_idx],meta_data,\
                                                          parse_from_single_pos_file)
                    if cut_images == 0:
                        img_array[pos_idx, ch_idx, t_idx] =\
                        parser.get_image(img_idx)
                    else:
                        img_array[pos_idx, ch_idx, t_idx] =\
                        parser.get_image(img_idx)[cut_images:-cut_images,\
                                                  cut_images:-cut_images]
    return img_array



def load_series_of_tifs(folder, channel=None):
    """
    function to load images/masks into a numpy array from a folder containing
    .tif images.

    Parameters
    ----------
    folder: str
        path to folder containing the .tif images relative to home dir
    positions: str
        if given, filter for images containing this string to filter
        for a certain channel.

    Returns
    -------
    all_im: np.array
        numpy array of dim (n_images, shape_images) containing all images
        of the folder

    """
    if channel:
        files = [f for f in sorted(myutils.listdir(folder)) if channel in f]
    else:
        files = sorted(myutils.listdir(folder))
    all_im = []
    for f in files:
        im = Image.open(f'{folder}{f}')
        all_im.append(np.array(im))
    all_im = np.array(all_im)
    return all_im


def load_masks_from_matlab(mask_filename, gt_folder):
    """
    function to load masks into a numpy array from a matlab file (.mat)

    Parameters
    ----------
    mask_filename: str
        filename of the matlab file which should be loaded
    gt_folder: str
        folder containing the matlab file

    Returns
    -------
    all_masks: np.array
        numpy array of dim (n_frames, shape_masks) containing all masks
        of the mat file

    """
    mask_fn_full = f'{gt_folder}{mask_filename}'
    mat = sio.loadmat(mask_fn_full)
    masks = mat['all_obj']['cells'][0][0]
    all_masks = np.moveaxis(masks, -1, 0)
    return all_masks



import sys
import os
import re
import traceback
import time
import datetime
import numpy as np
import pandas as pd

import skimage.io
from tifffile.tifffile import TiffWriter, TiffFile

from tqdm import tqdm

from PyQt5.QtWidgets import (
    QApplication, QMainWindow, QFileDialog,
    QVBoxLayout, QPushButton, QLabel, QMessageBox,
    QWidget
)
from PyQt5.QtCore import Qt, QEventLoop
from PyQt5 import QtGui

script_path = os.path.dirname(os.path.realpath(__file__))
cellacdc_path = os.path.join(os.path.dirname(script_path))
sys.path.append(cellacdc_path)

# Custom modules
from .. import prompts, load, myutils, apps, html_utils, widgets

from .. import qrc_resources

if os.name == 'nt':
    try:
        # Set taskbar icon in windows
        import ctypes
        myappid = 'schmollerlab.cellacdc.pyqt.v1' # arbitrary string
        ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(myappid)
    except Exception as e:
        pass

class alignWin(QMainWindow):
    def __init__(self, parent=None, allowExit=False,
                 actionToEnable=None, mainWin=None):
        self.allowExit = allowExit
        self.processFinished = False
        self.actionToEnable = actionToEnable
        self.mainWin = mainWin
        super().__init__(parent)
        self.setWindowTitle("Cell-ACDC - Align")
        self.setWindowIcon(QtGui.QIcon(":assign-motherbud.svg"))

        mainContainer = QWidget()
        self.setCentralWidget(mainContainer)

        mainLayout = QVBoxLayout()

        titleText = html_utils.paragraph(
            '<br><b>Alignment routine running...</b>', font_size='14px'
        )
        titleLabel = QLabel(titleText)
        mainLayout.addWidget(titleLabel)

        infoTxt = (
            'Follow the instructions in the pop-up windows.<br>'
            'Note that pop-ups might be minimized or behind other open windows.<br><br>'
            'Progess is displayed in the terminal/console.'
        )

        informativeLabel = QLabel(html_utils.paragraph(infoTxt))

        informativeLabel.setStyleSheet("padding:5px 0px 10px 0px;")
        # informativeLabel.setWordWrap(True)
        informativeLabel.setAlignment(Qt.AlignLeft)
        mainLayout.addWidget(informativeLabel)

        abortButton = QPushButton('Abort process')
        abortButton.clicked.connect(self.close)
        mainLayout.addWidget(abortButton)

        mainLayout.setContentsMargins(20, 0, 20, 20)
        mainContainer.setLayout(mainLayout)

    def getMostRecentPath(self):
        recentPaths_path = os.path.join(
            cellacdc_path, 'temp', 'recentPaths.csv'
        )
        if os.path.exists(recentPaths_path):
            df = pd.read_csv(recentPaths_path, index_col='index')
            if 'opened_last_on' in df.columns:
                df = df.sort_values('opened_last_on', ascending=False)
            self.MostRecentPath = df.iloc[0]['path']
            if not isinstance(self.MostRecentPath, str):
                self.MostRecentPath = ''
        else:
            self.MostRecentPath = ''

    def main(self):
        self.getMostRecentPath()
        exp_path = QFileDialog.getExistingDirectory(
            self, 'Select experiment folder containing Position_n folders '
                  'or specific Position_n folder', self.MostRecentPath)
        self.addToRecentPaths(exp_path)

        if exp_path == '':
            abort = self.doAbort()
            if abort:
                self.close()
                return

        self.setWindowTitle(f'Cell-ACDC - Align - "{exp_path}"')

        if os.path.basename(exp_path).find('Position_') != -1:
            is_pos_folder = True
        else:
            is_pos_folder = False

        if os.path.basename(exp_path).find('Images') != -1:
            is_images_folder = True
        else:
            is_images_folder = False

        print('Loading data...')

        if not is_pos_folder and not is_images_folder:
            select_folder = load.select_exp_folder()
            values = select_folder.get_values_segmGUI(exp_path)
            if not values:
                txt = (
                    'The selected folder:\n\n '
                    f'{exp_path}\n\n'
                    'is not a valid folder. '
                    'Select a folder that contains the Position_n folders'
                )
                msg = QMessageBox()
                msg.critical(
                    self, 'Incompatible folder', txt, msg.Ok
                )
                self.close()
                return


            select_folder.QtPrompt(self, values, allow_abort=False, show=True)
            if select_folder.was_aborted:
                abort = self.doAbort()
                if abort:
                    self.close()
                    return


            pos_foldernames = select_folder.selected_pos
            images_paths = [os.path.join(exp_path, pos, 'Images')
                            for pos in pos_foldernames]

        elif is_pos_folder:
            pos_foldername = os.path.basename(exp_path)
            exp_path = os.path.dirname(exp_path)
            images_paths = [f'{exp_path}/{pos_foldername}/Images']

        elif is_images_folder:
            images_paths = [exp_path]

        proceed, selectedFilenames = self.selectFiles(
            images_paths[0], filterExt=['npz', 'npy', 'tif']
        )
        if not proceed:
            abort = self.doAbort()
            if abort:
                self.close()
                return

        revertAlignment = self.askAlignmentMode()
        if revertAlignment is None:
            abort = self.doAbort()
            if abort:
                self.close()
                return

        abort, appendTxts = self.askTxtAppend()
        if abort:
            abort = self.doAbort()
            if abort:
                self.close()
                return

        appendedTxt = appendTxts[0]

        print('Aligning data...')
        if len(selectedFilenames) > 1:
            ch_name_selector = prompts.select_channel_name()
            channelNames, abort = ch_name_selector.get_available_channels(
                selectedFilenames, images_paths[0], useExt=None
            )
            if abort or not channelNames:
                self.criticalNoCommonBasename(
                    selectedFilenames, images_paths[0]
                )
                self.close()
                return

            self.prevSizeT, self.prevSizeZ = 1, 1
            for images_path in tqdm(images_paths, ncols=100):
                for chName in channelNames:
                    if chName.find('align_shift.npy') != -1:
                        continue
                    shifts, shifts_found = load.load_shifts(images_path)
                    if not shifts_found:
                        print('')
                        print('=============================')
                        print('WARNING: "Align_shift.npy" not found in folder '
                              f'{images_path}. Skipping it')
                        print('=============================')
                        continue
                    filenames = myutils.listdir(images_path)
                    chNameFile = [f for f in filenames if f.find(f'{chName}.')!=-1]
                    if not chNameFile:
                        print('')
                        print('=============================')
                        print(f'WARNING: File ending with "{chName}." not found in folder '
                              f'{images_path}. Skipping it')
                        continue

                    filePath = os.path.join(images_path, chNameFile[0])
                    alignedData = self.loadAndAlign(
                        filePath, shifts, revertAlignment
                    )
                    self.save(alignedData, filePath, appendedTxt)

        else:
            for images_path in tqdm(images_paths, ncols=100):
                shifts, shifts_found = load.load_shifts(images_path)
                if not shifts_found:
                    print('')
                    print('=============================')
                    print('WARNING: "Align_shift.npy" not found in folder '
                          f'{images_path}. Skipping it')
                    print('=============================')
                    continue
                # print(f'Aligning {filePath}...')
                filePath = os.path.join(images_path, selectedFilenames[0])
                alignedData = self.loadAndAlign(
                    filePath, shifts, revertAlignment
                )
                self.save(alignedData, filePath, appendedTxt)

        self.alignmentDone(filePath, appendedTxt)
        self.close()
        if self.allowExit:
            exit('Done.')

    def alignmentDone(self, filePath, appendedTxt):
        msg = widgets.myMessageBox()
        msg.setWidth(700)
        parent_path = os.path.dirname(filePath)
        filename, ext = os.path.splitext(os.path.basename(filePath))
        path = os.path.join(parent_path, f'{filename}_{appendedTxt}{ext}')
        txt = (
            'Alignment routine <b>ended successfully</b>.<br><br>'
            'File(s) saved to:<br><br>'
            f'<code>{path}</code>'
        )
        msg.addShowInFileManagerButton(parent_path)
        msg.information(
            self, 'Alignment done!', html_utils.paragraph(txt)
        )

    def save(self, alignedData, filePath, appendedTxt, first_call=True):
        dir = os.path.dirname(filePath)
        filename, ext = os.path.splitext(os.path.basename(filePath))
        path = os.path.join(dir, f'{filename}_{appendedTxt}{ext}')
        if ext == '.npz':
            np.savez_compressed(path, alignedData)
        elif ext == '.npy':
            np.save(path, alignedData)
        elif ext == '.tif':
            with TiffFile(filePath) as tif:
                metadata = tif.imagej_metadata
            try:
                info = metadata['Info']
                SizeT, SizeZ = self.readSizeTZ(info)
            except Exception as e:
                SizeT = len(alignedData)
                SizeZ = 1
            if SizeT != self.prevSizeT or SizeZ != self.prevSizeZ:
                SizeT, SizeZ = self.askImageSizeZT(SizeT, SizeZ)
            myutils.imagej_tiffwriter(path, alignedData, metadata, SizeT, SizeZ)

    def readSizeTZ(self, info):
        SizeT = int(re.findall('SizeT = (\d+)', info)[0])
        SizeZ = int(re.findall('SizeZ = (\d+)', info)[0])
        return SizeT, SizeZ

    def askImageSizeZT(self, SizeT, SizeZ):
        font = QtGui.QFont()
        font.setPixelSize(13)
        win = apps.QDialogAcdcInputs(SizeT, SizeZ, None, None,
                                     show_finterval=False,
                                     parent=self, font=font)
        win.setFont(font)
        win.show()
        win.setWidths(font=font)
        win.exec_()
        return win.SizeT, win.SizeZ

    def askTxtAppend(self):
        font = QtGui.QFont()
        font.setPixelSize(13)
        self.win = apps.QDialogEntriesWidget(
            winTitle='Appended name',
            entriesLabels=['Type a name to append at the end of each aligned file:'],
            defaultTxts=['aligned'],
            parent=self, font=font
        )
        self.win.exec_()
        return self.win.cancel, self.win.entriesTxt

    def loadAndAlign(self, filePath, shifts, revertAlignment):
        filename = os.path.basename(filePath)
        _, ext = os.path.splitext(filename)
        if ext == '.npz':
            data = np.load(filePath)['arr_0']
        elif ext == '.npy':
            data = np.load(filePath)
        elif ext == '.tif':
            data = skimage.io.imread(filePath)
        if revertAlignment:
            shifts = -shifts
        alignedData = np.zeros_like(data)
        for frame_i, shift in enumerate(tqdm(shifts, leave=False)):
            if frame_i >= len(data):
                break
            img = data[frame_i]
            axis = tuple(range(img.ndim))[-2:]
            aligned_img = np.roll(img, tuple(shift), axis=axis)
            alignedData[frame_i] = aligned_img
        return alignedData

    def criticalNoCommonBasename(self, filenames, parent_path):
        myutils.checkDataIntegrity(filenames, parent_path, parentQWidget=self)

    def askAlignmentMode(self):
        msg = widgets.myMessageBox()
        txt = (
            'Do you want to <b>revert</b> a previously applied <b>alignment</b> '
            'or <b>apply alignment</b>?'
        )
        _, revertButton, applyButton = msg.question(
            self, 'Alignment mode', html_utils.paragraph(txt),
            buttonsTexts=('Cancel', 'Revert alignment', 'Apply alignment')
        )
        if msg.cancel:
            return None
        if msg.clickedButton == revertButton:
            revertAlignment = True
        else:
            revertAlignment = False
        return revertAlignment

    def selectFiles(self, images_path, filterExt=None):
        files = myutils.listdir(images_path)
        if filterExt is not None:
            items = []
            for file in files:
                _, ext = os.path.splitext(file)
                for allowedExt in filterExt:
                    if ext.find(allowedExt) != -1:
                        items.append(file)
        else:
            items = files

        selectFilesWidget = apps.QDialogListbox(
            'Select files',
            'Select to which files you want to apply alignment\n\n'
            'NOTE: if you selected multiple Position folders I will try \n'
            'to apply alignment to all the selected files in each Position folder',
            items, multiSelection=True, parent=self
        )
        selectFilesWidget.exec_()

        if selectFilesWidget.cancel:
            return False, []

        selectedFilenames = selectFilesWidget.selectedItemsText
        if not selectedFilenames:
            return False, []
        else:
            return True, selectedFilenames

    def addToRecentPaths(self, exp_path):
        if not os.path.exists(exp_path):
            return
        recentPaths_path = os.path.join(
            cellacdc_path, 'temp', 'recentPaths.csv'
        )
        if os.path.exists(recentPaths_path):
            df = pd.read_csv(recentPaths_path, index_col='index')
            recentPaths = df['path'].to_list()
            if 'opened_last_on' in df.columns:
                openedOn = df['opened_last_on'].to_list()
            else:
                openedOn = [np.nan]*len(recentPaths)
            if exp_path in recentPaths:
                pop_idx = recentPaths.index(exp_path)
                recentPaths.pop(pop_idx)
                openedOn.pop(pop_idx)
            recentPaths.insert(0, exp_path)
            openedOn.insert(0, datetime.datetime.now())
            # Keep max 20 recent paths
            if len(recentPaths) > 20:
                recentPaths.pop(-1)
                openedOn.pop(-1)
        else:
            recentPaths = [exp_path]
            openedOn = [datetime.datetime.now()]
        df = pd.DataFrame({'path': recentPaths,
                           'opened_last_on': pd.Series(openedOn,
                                                       dtype='datetime64[ns]')})
        df.index.name = 'index'
        df.to_csv(recentPaths_path)

    def doAbort(self):
        if self.allowExit:
            exit('Execution aborted by the user')
        else:
            print('Segmentation routine aborted by the user.')
            return True

    def closeEvent(self, event):
        if self.actionToEnable is not None:
            self.actionToEnable.setDisabled(False)
            self.mainWin.setWindowState(Qt.WindowNoState)
            self.mainWin.setWindowState(Qt.WindowActive)
            self.mainWin.raise_()


if __name__ == "__main__":
    print('Launching alignment script...')
    # Handle high resolution displays:
    if hasattr(Qt, 'AA_EnableHighDpiScaling'):
        QApplication.setAttribute(Qt.AA_EnableHighDpiScaling, True)
    if hasattr(Qt, 'AA_UseHighDpiPixmaps'):
        QApplication.setAttribute(Qt.AA_UseHighDpiPixmaps, True)
    # Create the application
    app = QApplication(sys.argv)
    app.setStyle(QtGui.QStyleFactory.create('Fusion'))
    win = alignWin(allowExit=True)
    win.show()
    print('Done. If window asking to select a folder is not visible, it is '
          'behind some other open window.')
    win.main()
    sys.exit(app.exec_())



# Python-bioformats is distributed under the GNU General Public
# License, but this file is licensed under the more permissive BSD
# license.  See the accompanying file LICENSE for details.
#
# Copyright (c) 2009-2014 Broad Institute
# All rights reserved.

import javabridge

def basic_config():
    '''Configure logging for "WARN" level'''
    log4j = javabridge.JClassWrapper("loci.common.Log4jTools")
    log4j.enableLogging()
    log4j.setRootLevel("WARN")



import os
import sys

import matplotlib.pyplot as plt

from collections import namedtuple
from importlib import import_module

from . import apps, myutils

from PyQt5.QtWidgets import QApplication, QStyleFactory

import pytest
pytest.skip('skipping this test since it is gui based', allow_module_level=True)

# Ask which model to use --> Test if new model is visible
app = QApplication(sys.argv)
app.setStyle(QStyleFactory.create('Fusion'))


cellacdc_path = os.path.dirname(os.path.abspath(__file__))
models = myutils.listdir(os.path.join(cellacdc_path, 'models'))
win = apps.QDialogListbox(
    'Select model',
    'Select model to use for segmentation: ',
    models,
    multiSelection=False
)
win.exec_()

if win.cancel:
    sys.exit('Execution aborted')

# Load model as a module
model_name = win.selectedItemsText[0]
acdcSegment = import_module(f'models.{model_name}.acdcSegment')

# Read all models parameters
init_params, segment_params = myutils.getModelArgSpec(acdcSegment)

# Prompt user to enter the model parameters
try:
    url = acdcSegment.url_help()
except AttributeError:
    url = None

win = apps.QDialogModelParams(
    init_params,
    segment_params,
    model_name, url=url)

win.exec_()

# Initialize model
model = acdcSegment.Model(**win.init_kwargs)

# Use model on a test image
# In this case image is in 'Cell-ACDC/data/test_images' folder
import skimage.io

main_path = os.path.dirname(cellacdc_path)
test_images_path = os.path.join(main_path, 'data', 'test_images')

test_img_path = os.path.join(test_images_path, 'test_YeaZ.tif')

img = skimage.io.imread(test_img_path)

print(img.shape)

lab = model.segment(img, **win.segment2D_kwargs)

print(lab.shape)

if model_name == 'YeastMate':
    cca_df = model.predictCcaState(img)
    print(cca_df)

fig, ax = plt.subplots(1, 2)
ax[0].imshow(img)
ax[1].imshow(lab)
plt.show()



import numpy as np
import pathlib
import re
import sys
import os
import traceback
import shutil
from importlib import import_module

from skimage.measure._regionprops import PROPS, COL_DTYPES, RegionProperties

from . import core, base_cca_df, html_utils

user_path = pathlib.Path.home()
acdc_metrics_path = os.path.join(user_path, 'acdc-metrics')
if not os.path.exists(acdc_metrics_path):
    os.makedirs(acdc_metrics_path)
sys.path.append(acdc_metrics_path)

cellacdc_path = os.path.dirname(os.path.abspath(__file__))
metrics_path = os.path.join(cellacdc_path, 'metrics')

# Copy metrics to acdc-metrics user path
for file in os.listdir(metrics_path):
    if not file.endswith('.py'):
        continue
    src = os.path.join(metrics_path, file)
    dst = os.path.join(acdc_metrics_path, file)
    shutil.copy(src, dst)

def get_all_metrics_names():
    all_metrics_names = []
    custom_metrics_names = list(_get_custom_metrics_names().keys())
    size_metrics_names = list(get_size_metrics_desc().keys())
    standard_metrics_names = list(_get_metrics_names().keys())
    bkgr_val_names = list(_get_bkgr_val_names().keys())
    props_names = get_props_names()
    all_metrics_names.extend(custom_metrics_names)
    all_metrics_names.extend(size_metrics_names)
    all_metrics_names.extend(standard_metrics_names)
    all_metrics_names.extend(bkgr_val_names)
    all_metrics_names.extend(props_names)
    return all_metrics_names

def get_all_acdc_df_colnames():
    all_acdc_df_colnames = get_all_metrics_names()
    all_acdc_df_colnames.append('frame_i')
    all_acdc_df_colnames.append('time_seconds')
    all_acdc_df_colnames.append('Cell_ID')
    cca_df_colnames = list(base_cca_df.keys())
    all_acdc_df_colnames.extend(cca_df_colnames)
    additional_colnames = [
        'is_cell_dead',
        'is_cell_excluded',
        'x_centroid',
        'y_centroid',
        'editIDclicked_x',
        'editIDclicked_y',
        'editIDnewID',
        'editIDnewIDs'
    ]
    all_acdc_df_colnames.extend(additional_colnames)
    return all_acdc_df_colnames


def get_custom_metrics_func():
    scripts = os.listdir(acdc_metrics_path)
    custom_func_dict = {}
    for file in scripts:
        module_name, ext = os.path.splitext(file)
        if ext != '.py':
            # print(f'The file {file} is not a python file. Ignoring it.')
            continue
        try:
            module = import_module(module_name)
            func = getattr(module, module_name)
            custom_func_dict[module_name] = func
        except Exception:
            traceback.print_exc()
    return custom_func_dict

def _get_custom_metrics_names():
    custom_func_dict = get_custom_metrics_func()
    keys = custom_func_dict.keys()
    custom_metrics_names = {func_name:func_name for func_name in keys}
    return custom_metrics_names

def custom_metrics_desc(isZstack, chName):
    how_3Dto2D = ['_maxProj', '_meanProj', '_zSlice'] if isZstack else ['']
    how_3Dto2D_desc = _get_how_3Dto2D_desc()
    custom_metrics_names = _get_custom_metrics_names()
    custom_metrics_desc = {}
    for how, how_desc in zip(how_3Dto2D, how_3Dto2D_desc):
        for func_name, func_desc in custom_metrics_names.items():
            metric_name = f'{chName}_{func_name}{how}'
            if isZstack:
                note_txt = html_utils.paragraph(f"""
                    {_get_zStack_note(how_desc)}
                    Example: <code>{metric_name}</code> is the
                    <b>{func_desc.lower()}</b> of the {chName} signal after
                    converting 3D to 2D {how_desc}
                """)
            else:
                note_txt = ''

            desc = html_utils.paragraph(f"""
                <b>{func_desc}</b> is a custom defined measurement.<br>
                {note_txt}
            """)
            custom_metrics_desc[metric_name] = desc
    return custom_metrics_desc

def _um3():
    return '<code>&micro;m<sup>3</sup></code>'

def _um2():
    return '<code>&micro;m<sup>2</sup></code>'

def _fl():
    return '<code>fl</code>'

def _get_zStack_note(how_desc):
    s = (f"""
        <i>NOTE: since you loaded <b>3D z-stacks</b>, Cell-ACDC needs
        to convert the z-stacks to 2D images {how_desc} for this metric.<br>
        This is specified in the name of the column.<br><br></i>
    """)
    return s

def _get_how_3Dto2D_desc():
    how_3Dto2D_desc = [
        'using a <b>max projection</b>',
        'using a <b>mean projection</b> (recommended for <b>confocal imaging</b>)',
        'using the <b>z-slice you used for segmentation</b> '
        '(recommended for <b>epifluorescence imaging</b>)'
    ]
    return how_3Dto2D_desc

def get_size_metrics_desc():
    url = 'https://www.nature.com/articles/s41467-020-16764-x#Sec16'
    size_metrics = {
        'cell_area_pxl': html_utils.paragraph("""
            <b>Area of the segmented object</b> in pixels, i.e.,
            total number of pixels in the object.
        """),
        'cell_vol_vox': html_utils.paragraph(f"""
            <b>Estimated volume of the segmented object</b> in voxels.<br><br><br>
            To calculate object volume based on 2D masks, the object is first
            <b>aligned along its major axis</b>.<br><br>
            Next, it is divided into slices perpendicular to the
            major axis, where the height of each slice is one pixel
            (i.e., one row).<br><br>
            We then assume <b>rotational symmetry</b> of each slice around
            its middle axis parallel to the object's major axis.<br><br>
            For each slice Cell-ACDC calculates the volume of the resulting
            <b>cylinder</b> with one pixel height and diameter the width
            of the respective slice.<br><br>
            Finally, the volumes of each cylinder are added to obtain
            total object volume.<br><br>
            <i> Note that in <a href=\"{url}">this</a> publication we
            showed that this method strongly correlates with volume
            computed from a 3D segmentation mask.</i>
        """),
        'cell_area_um2': html_utils.paragraph(f"""
            <b>Area of the segmented object</b> in {_um2()}, i.e.,
            total number of pixels in the object.<br><br>
            Conversion from pixels to {_um2()} is perfomed using
            the provided pixel size.
        """),
        'cell_vol_fl': html_utils.paragraph(f"""
            <b>Estimated volume of the segmented object</b> in {_um3()}.<br><br><br>

            To calculate object volume based on 2D masks, the object is first
            <b>aligned along its major axis</b>.<br><br>

            Next, it is divided into slices perpendicular to the
            major axis, where the height of each slice is one pixel
            (i.e., one row).<br><br>

            We then assume <b>rotational symmetry</b> of each slice around
            its middle axis parallel to the object's major axis.<br><br>

            For each slice Cell-ACDC calculates the volume of the resulting
            <b>cylinder</b> with one pixel height and diameter the width
            of the respective slice.<br><br>

            Finally, the volumes of each cylinder are added and converted
            to {_fl()} (same as {_um3()}) using the provided pixel size
            to obtain total object volume.<br><br>

            <i> Note that in <a href=\"{url}">this</a> publication we
            showed that this method strongly correlates with volume
            computed from a 3D segmentation mask.</i>
        """)
    }
    return size_metrics

def standard_metrics_desc(isZstack, chName):
    how_3Dto2D = ['_maxProj', '_meanProj', '_zSlice'] if isZstack else ['']
    how_3Dto2D_desc = _get_how_3Dto2D_desc()
    metrics_names = _get_metrics_names()
    bkgr_val_names = _get_bkgr_val_names()
    metrics_desc = {}
    bkgr_val_desc = {}
    for how, how_desc in zip(how_3Dto2D, how_3Dto2D_desc):
        for func_name, func_desc in metrics_names.items():
            metric_name = f'{chName}_{func_name}{how}'
            if isZstack:
                note_txt = (f"""
                {_get_zStack_note(how_desc)}
                Example: <code>{metric_name}</code> is the
                <b>{func_desc.lower()}</b> of the {chName} signal after
                converting 3D to 2D {how_desc}
                """)
            else:
                note_txt = ''

            if func_desc == 'Amount':
                amount_desc = ("""
                    Amount is the <b>background corrected total
                    fluorescent intensity</b>, which is usually the best proxy
                    for the amount of the tagged molecule, e.g.,
                    <b>protein amount</b>.
                    <br><br>
                """)
                main_desc = f'<b>{func_desc}</b> computed from'
            elif func_desc == 'Concentration':
                amount_desc = ("""
                    Concentration is given by <code>Amount/cell_volume</code>,
                    where amount is the <b>background corrected total
                    fluorescent intensity</b>. Amount is usually the best proxy
                    for the amount of the tagged molecule, e.g.,
                    <b>protein amount</b>.
                    <br><br>
                """)
                main_desc = f'<b>{func_desc}</b> computed from'
            else:
                amount_desc = ''
                main_desc = f'<b>{func_desc}</b> computed from'

            if func_name == 'amount_autoBkgr':
                bkgr_desc = ("""
                    <code>autoBkgr</code> means that the background value
                    used to correct the intensities is computed as the <b>median</b>
                    of ALL the pixels outside of the segmented objects
                    (i.e., pixels with ID 0 in the segmentation mask)
                    <br><br>
                """)
            elif func_name == 'amount_dataPrepBkgr':
                bkgr_desc = ("""
                    <code>dataPrepBkgr</code> means that the background value
                    used to correct the intensities is computed as the <b>median</b>
                    of the pixels from the pixels inside the rectangular
                    <b>background ROIs</b> that you selected in the
                    data prep module (module 1.).<br>
                    If you didn't do this step, then the amount will be set to 0.
                    <br><br>
                """)
            else:
                bkgr_desc = ''

            desc = html_utils.paragraph(f"""
                {main_desc} the pixels inside
                each segmented object.<br><br>
                {amount_desc}{bkgr_desc}{note_txt}
            """)
            metrics_desc[metric_name] = desc

        for bkgr_name, bkgr_desc in bkgr_val_names.items():
            bkgr_colname = f'{chName}_{bkgr_name}{how}'
            if isZstack:
                note_txt = (f"""
                {_get_zStack_note(how_desc)}
                Example: <code>{bkgr_colname}</code> is the
                <b>{bkgr_desc.lower()}</b> of the {chName} background after
                converting 3D to 2D {how_desc}
                """)
            else:
                note_txt = ''

            if bkgr_name.find('autoBkgr') != -1:
                bkgr_type_desc = ("""
                    <code>autoBkgr</code> means that the background value
                    is computed from ALL the pixels outside
                    of the segmented objects
                    (i.e., pixels with ID 0 in the segmentation mask)
                    <br><br>
                """)
            else:
                bkgr_type_desc = ("""
                    <code>dataPrepBkgr</code> means that the background value
                    is computed from the pixels inside the rectangular
                    <b>background ROIs</b> that you selected in the
                    data prep module (module 1.).<br>
                    If you didn't do this step, then this value will be set to 0.
                    <br><br>
                """)

            bkgr_final_desc = html_utils.paragraph(f"""
                <b>{bkgr_desc}</b> of the background intensities.<br><br>
                {bkgr_type_desc}{note_txt}
            """)
            bkgr_val_desc[bkgr_colname] = bkgr_final_desc

    return metrics_desc, bkgr_val_desc

def get_conc_keys(amount_colname):
    conc_key_vox = re.sub(
        r'amount_([A-Za-z]+)',
        r'concentration_\1_from_vol_vox',
        amount_colname
    )
    conc_key_fl = conc_key_vox.replace('from_vol_vox', 'from_vol_fl')
    return conc_key_vox, conc_key_fl

def _get_metrics_names():
    metrics_names = {
        'mean': 'Mean',
        'sum': 'Sum',
        'amount_autoBkgr': 'Amount',
        'amount_dataPrepBkgr': 'Amount',
        'concentration_autoBkgr_from_vol_vox': 'Concentration',
        'concentration_dataPrepBkgr_from_vol_vox': 'Concentration',
        'concentration_autoBkgr_from_vol_fl': 'Concentration',
        'concentration_dataPrepBkgr_from_vol_fl': 'Concentration',
        'median': 'Median',
        'min': 'Minimum',
        'max': 'Maximum',
        'q25': '25 percentile',
        'q75': '75 percentile',
        'q05': '5 percentile',
        'q95': '95 percentile'
    }
    return metrics_names

def _get_bkgr_val_names():
    bkgr_val_names = {
        'autoBkgr_bkgrVal_median': 'Median',
        'autoBkgr_bkgrVal_mean': 'Mean',
        'autoBkgr_bkgrVal_q75': '75 percentile',
        'autoBkgr_bkgrVal_q25': '25 percentile',
        'autoBkgr_bkgrVal_q95': '95 percentile',
        'autoBkgr_bkgrVal_q05': '5 percentile',
        'dataPrepBkgr_bkgrVal_median': 'Median',
        'dataPrepBkgr_bkgrVal_mean': 'Mean',
        'dataPrepBkgr_bkgrVal_q75': '75 percentile',
        'dataPrepBkgr_bkgrVal_q25': '25 percentile',
        'dataPrepBkgr_bkgrVal_q95': '95 percentile',
        'dataPrepBkgr_bkgrVal_q05': '5 percentile',
    }
    return bkgr_val_names

def get_props_info_txt():
    url = 'https://scikit-image.org/docs/0.18.x/api/skimage.measure.html#skimage.measure.regionprops'
    txt = html_utils.paragraph(f"""
        Morphological properties are calculated using the function
        from the library <code>scikit-image</code> called
        <code>skimage.measure.regionprops</code>.<br><br>
        You can find more details about each one of the properties
        <a href=\"{url}"></a>.
    """)
    return txt

def _is_numeric_dtype(dtype):
    is_numeric = (
        dtype is float or dtype is int
    )
    return is_numeric

def get_props_names():
    props = (
        'label',
        'inertia_tensor_eigvals',
        'major_axis_length',
        'equivalent_diameter',
        'moments',
        'area',
        'solidity',
        'feret_diameter_max',
        'extent',
        'inertia_tensor',
        'filled_area',
        'centroid',
        'bbox_area',
        'local_centroid',
        'convex_area',
        'euler_number',
        'minor_axis_length',
        'moments_normalized',
        'moments_central',
        'bbox'
    )
    return props

def standard_metrics_func():
    metrics_func = {
        'mean': lambda arr: arr.mean(),
        'sum': lambda arr: arr.sum(),
        'amount_autoBkgr': lambda arr, bkgr, area: (arr.mean()-bkgr)*area,
        'amount_dataPrepBkgr': lambda arr, bkgr, area: (arr.mean()-bkgr)*area,
        'median': lambda arr: np.median(arr),
        'min': lambda arr: np.min(arr),
        'max': lambda arr: np.max(arr),
        'q25': lambda arr: np.quantile(arr, q=0.25),
        'q75': lambda arr: np.quantile(arr, q=0.75),
        'q05': lambda arr: np.quantile(arr, q=0.05),
        'q95': lambda arr: np.quantile(arr, q=0.95)
    }
    all_metrics_names = list(_get_metrics_names().keys())

    bkgr_val_names = tuple(_get_bkgr_val_names().keys())
    all_metrics_names.extend(bkgr_val_names)

    return metrics_func, all_metrics_names

def add_metrics_instructions():
    url = 'https://github.com/SchmollerLab/Cell_ACDC/issues'
    href = f'<a href="{url}">here</a>'
    s = html_utils.paragraph(f"""
    To add custom metrics to the <code>acdc_output.csv</code>
    file you need to <b>create a python script and save it into the
    following folder</b>
    <code>{acdc_metrics_path}</code><br><br>
    Name the file with the name of the column you want to add. Inside the
    file, create a <b>function with the same name of the file</b>
    similar to the example below:<br><br>
    Pseudo-code:
    <pre><code>
    def <b>CV</b>(signal, autoBkgr, dataPrepBkgr, correct_with_bkgr=False, which_bkgr='auto'):
        if correct_with_bkgr:
            if which_bkgr=='auto':
                signal = signal - autoBkgr
            elif dataPrepBkgr is not None:
                signal = signal - dataPrepBkgr

        <i># Here goes your custom metric computation</i>
        CV = np.std(signal)/np.mean(signal)

        return CV
    </code></pre>
    Have a look at the <code>CV.py</code> file (click on "Show example..." below)
    for a full example.<br><br>
    <i>If it doesn't work, please report the issue {href} with the
    code you wrote. Thanks.</i>
    """)
    return s, metrics_path



from functools import wraps
import re

from . import is_mac

def _tag(tag_info='p style="font-size:10px"'):
    def wrapper(func):
        @wraps(func)
        def inner(text):
            tag = tag_info.split(' ')[0]
            text = f'<{tag_info}>{text}</{tag}>'
            return text
        return inner
    return wrapper

def tag(text, tag_info='p style="font-size:10pt"'):
    tag = tag_info.split(' ')[0]
    text = f'<{tag_info}>{text}</{tag}>'
    return text

def css_head(txt):
    # if is_mac:
    #     txt = txt.replace(',', ',&nbsp;')
    s = (f"""
    <!DOCTYPE html>
    <html>
    <head>
    <style type="text/css">
            {txt}
    </style>
    </head>
    """)
    return s

def html_body(txt):
    if is_mac:
        txt = txt.replace(',', ',&nbsp;')
    s = (f"""
    <body>
        {txt}
    </body>
    </html>
    """)
    return s

def paragraph(txt, font_size='13px', font_color=None, wrap=True):
    # if is_mac:
    #     # Qt < 5.15.3 has a bug on macOS and the space after comma and perdiod
    #     # are super small. Force a non-breaking space (except for 'e.g.,').
    #     txt = txt.replace(',', ',&nbsp;')
    #     txt = txt.replace('.', '.&nbsp;')
    #     txt = txt.replace('e.&nbsp;g.&nbsp;', 'e.g.')
    #     txt = txt.replace('.&nbsp;.&nbsp;.&nbsp;', '...')
    #     txt = txt.replace('i.&nbsp;e.&nbsp;', 'i.e.')
    #     txt = txt.replace('etc.&nbsp;)', 'etc.)')
    if not wrap:
        txt = txt.replace(' ', '&nbsp;')
    if font_color is None:
        s = (f"""
        <p style="font-size:{font_size};">
            {txt}
        </p>
        """)
    else:
        s = (f"""
        <p style="font-size:{font_size}; color:{font_color}">
            {txt}
        </p>
        """)
    return s



# Test importing the cellacdc modules.

import cellacdc

from cellacdc import gui
from cellacdc import dataStruct
from cellacdc import segm
from cellacdc import dataPrep

def test_placeholder():
    # Add test test_placeholder because we are only testing import
    pass



import os
import javabridge as jv

from cellacdc import bioformats

# import bioformats

print(bioformats.__file__)

# path = r'"G:\My Drive\01_Postdoc_HMGU\Python_MyScripts\MIA\Git\Cell_ACDC\cellacdc\bioformats\jars\bioformats_package.jar"'
jars = bioformats.JARS
print(jars)
jv.start_vm(class_path=jars)
paths = jv.JClassWrapper('java.lang.System').getProperty('java.class.path').split(";")

for path in paths:
    print("%s: %s" %("exists" if os.path.isfile(path) else "missing", path))

jv.kill_vm()



from stardist.models import StarDist2D
from csbdeep.utils import normalize
import numpy as np
import os

from cellacdc import models

class Model:
    def __init__(self, model_name='T_cell'):
        stardist_default_models = models.STARDIST_MODELS
        stardist_path = os.path.dirname(os.path.abspath(__file__))
        T_cell_path = os.path.join(stardist_path, 'model', 'T_cell')
        if not os.path.exists(T_cell_path):
            model_name = stardist_default_models[0]
        # Initialize model
        if model_name in stardist_default_models:
            self.model = StarDist2D.from_pretrained(model_name)
        else:
            script_path = os.path.abspath(__file__)
            stardist_path = os.path.dirname(script_path)
            model_path = os.path.join(stardist_path, 'model')
            self.model = StarDist2D(None, name=model_name, basedir=model_path)

    def segment(self, image, prob_thresh=0.0, nms_thresh=0.0):
        # Preprocess image
        prob_thresh = prob_thresh if prob_thresh > 0 else None
        nms_thresh = nms_thresh if nms_thresh > 0 else None
        lab, _ = self.model.predict_instances(
            normalize(image),
            prob_thresh=prob_thresh,
            nms_thresh=nms_thresh
        )
        return lab.astype(np.uint16)
